@@ -124,7 +124,7 @@ struct _GcrParsed {
 	GckAttributes *attrs;
 	const gchar *description;
 	gchar *label;
-	EggBytes *data;
+	GBytes *data;
 	gboolean sensitive;
 	GcrDataFormat format;
 	struct _GcrParsed *next;
@@ -148,7 +148,7 @@ typedef struct {
 
 typedef struct _ParserFormat {
 	gint format_id;
-	gint (*function) (GcrParser *self, EggBytes *data);
+	gint (*function) (GcrParser *self, GBytes *data);
 } ParserFormat;
 
 /* Forward declarations */
@@ -228,12 +228,12 @@ parsed_attribute (GcrParsed *parsed,
 static void
 parsed_attribute_bytes (GcrParsed *parsed,
                         CK_ATTRIBUTE_TYPE type,
-                        EggBytes *data)
+                        GBytes *data)
 {
 	g_assert (parsed != NULL);
 	gck_builder_add_data (&parsed->builder, type,
-	                      egg_bytes_get_data (data),
-	                      egg_bytes_get_size (data));
+	                      g_bytes_get_data (data, NULL),
+	                      g_bytes_get_size (data));
 }
 
 static gboolean
@@ -242,7 +242,7 @@ parsed_asn1_number (GcrParsed *parsed,
                     const gchar *part,
                     CK_ATTRIBUTE_TYPE type)
 {
-	EggBytes *value;
+	GBytes *value;
 
 	g_assert (asn);
 	g_assert (parsed);
@@ -252,7 +252,7 @@ parsed_asn1_number (GcrParsed *parsed,
 		return FALSE;
 
 	parsed_attribute_bytes (parsed, type, value);
-	egg_bytes_unref (value);
+	g_bytes_unref (value);
 	return TRUE;
 }
 
@@ -262,7 +262,7 @@ parsed_asn1_element (GcrParsed *parsed,
                      const gchar *part,
                      CK_ATTRIBUTE_TYPE type)
 {
-	EggBytes *value;
+	GBytes *value;
 
 	g_assert (asn);
 	g_assert (parsed);
@@ -272,7 +272,7 @@ parsed_asn1_element (GcrParsed *parsed,
 		return FALSE;
 
 	parsed_attribute_bytes (parsed, type, value);
-	egg_bytes_unref (value);
+	g_bytes_unref (value);
 	return TRUE;
 }
 
@@ -298,7 +298,7 @@ parsed_boolean_attribute (GcrParsed *parsed,
 static void
 parsing_block (GcrParsed *parsed,
                gint format,
-               EggBytes *data)
+               GBytes *data)
 {
 	g_assert (parsed != NULL);
 	g_assert (data != NULL);
@@ -306,7 +306,7 @@ parsing_block (GcrParsed *parsed,
 	g_assert (parsed->data == NULL);
 
 	parsed->format = format;
-	parsed->data = egg_bytes_ref (data);
+	parsed->data = g_bytes_ref (data);
 }
 
 static void
@@ -397,7 +397,7 @@ pop_parsed (GcrParser *self,
 	gck_builder_clear (&parsed->builder);
 	gck_attributes_unref (parsed->attrs);
 	if (parsed->data)
-		egg_bytes_unref (parsed->data);
+		g_bytes_unref (parsed->data);
 	g_free (parsed->label);
 	g_free (parsed);
 }
@@ -472,7 +472,7 @@ parsed_fire (GcrParser *self,
 
 static gint
 parse_der_private_key_rsa (GcrParser *self,
-                           EggBytes *data)
+                           GBytes *data)
 {
 	gint res = GCR_ERROR_UNRECOGNIZED;
 	GNode *asn = NULL;
@@ -527,7 +527,7 @@ done:
 
 static gint
 parse_der_private_key_dsa (GcrParser *self,
-                           EggBytes *data)
+                           GBytes *data)
 {
 	gint ret = GCR_ERROR_UNRECOGNIZED;
 	GNode *asn = NULL;
@@ -565,8 +565,8 @@ done:
 
 static gint
 parse_der_private_key_dsa_parts (GcrParser *self,
-                                 EggBytes *keydata,
-                                 EggBytes *params)
+                                 GBytes *keydata,
+                                 GNode *params)
 {
 	gint ret = GCR_ERROR_UNRECOGNIZED;
 	GNode *asn_params = NULL;
@@ -575,7 +575,7 @@ parse_der_private_key_dsa_parts (GcrParser *self,
 
 	parsed = push_parsed (self, TRUE);
 
-	asn_params = egg_asn1x_create_and_decode (pk_asn1_tab, "DSAParameters", params);
+	asn_params = egg_asn1x_get_any_as (params, pk_asn1_tab, "DSAParameters");
 	asn_key = egg_asn1x_create_and_decode (pk_asn1_tab, "DSAPrivatePart", keydata);
 	if (!asn_params || !asn_key)
 		goto done;
@@ -610,7 +610,7 @@ done:
 
 static gint
 parse_der_private_key (GcrParser *self,
-                       EggBytes *data)
+                       GBytes *data)
 {
 	gint res;
 
@@ -628,8 +628,8 @@ parse_der_private_key (GcrParser *self,
 static gint
 handle_subject_public_key_rsa (GcrParser *self,
                                GcrParsed *parsed,
-                               EggBytes *key,
-                               EggBytes *params)
+                               GBytes *key,
+                               GNode *params)
 {
 	gint res = GCR_ERROR_FAILURE;
 	GNode *asn = NULL;
@@ -655,15 +655,15 @@ done:
 static gint
 handle_subject_public_key_dsa (GcrParser *self,
                                GcrParsed *parsed,
-                               EggBytes *key,
-                               EggBytes *params)
+                               GBytes *key,
+                               GNode *params)
 {
 	gint res = GCR_ERROR_FAILURE;
 	GNode *key_asn = NULL;
 	GNode *param_asn = NULL;
 
 	key_asn = egg_asn1x_create_and_decode (pk_asn1_tab, "DSAPublicPart", key);
-	param_asn = egg_asn1x_create_and_decode (pk_asn1_tab, "DSAParameters", params);
+	param_asn = egg_asn1x_get_any_as (params, pk_asn1_tab, "DSAParameters");
 
 	if (!key_asn || !param_asn)
 		goto done;
@@ -687,11 +687,11 @@ done:
 
 static gint
 parse_der_subject_public_key (GcrParser *self,
-                              EggBytes *data)
+                              GBytes *data)
 {
 	GcrParsed *parsed;
-	EggBytes *params;
-	EggBytes *key;
+	GNode *params;
+	GBytes *key;
 	GNode *asn = NULL;
 	GNode *node;
 	GQuark oid;
@@ -708,8 +708,7 @@ parse_der_subject_public_key (GcrParser *self,
 	node = egg_asn1x_node (asn, "algorithm", "algorithm", NULL);
 	oid = egg_asn1x_get_oid_as_quark (node);
 
-	node = egg_asn1x_node (asn, "algorithm", "parameters", NULL);
-	params = egg_asn1x_get_element_raw (node);
+	params = egg_asn1x_node (asn, "algorithm", "parameters", NULL);
 
 	node = egg_asn1x_node (asn, "subjectPublicKey", NULL);
 	key = egg_asn1x_get_bits_as_raw (node, &bits);
@@ -723,8 +722,7 @@ parse_der_subject_public_key (GcrParser *self,
 	else
 		ret = GCR_ERROR_UNRECOGNIZED;
 
-	egg_bytes_unref (key);
-	egg_bytes_unref (params);
+	g_bytes_unref (key);
 
 	if (ret == SUCCESS)
 		parsed_fire (self, parsed);
@@ -741,13 +739,13 @@ parse_der_subject_public_key (GcrParser *self,
 
 static gint
 parse_der_pkcs8_plain (GcrParser *self,
-                       EggBytes *data)
+                       GBytes *data)
 {
 	gint ret;
 	CK_KEY_TYPE key_type;
 	GQuark key_algo;
-	EggBytes *keydata = NULL;
-	EggBytes *params = NULL;
+	GBytes *keydata = NULL;
+	GNode *params = NULL;
 	GNode *asn = NULL;
 	GcrParsed *parsed;
 
@@ -775,11 +773,11 @@ parse_der_pkcs8_plain (GcrParser *self,
   		goto done;
   	}
 
-	keydata = egg_asn1x_get_raw_value (egg_asn1x_node (asn, "privateKey", NULL));
+	keydata = egg_asn1x_get_string_as_bytes (egg_asn1x_node (asn, "privateKey", NULL));
 	if (!keydata)
 		goto done;
 
-	params = egg_asn1x_get_element_raw (egg_asn1x_node (asn, "privateKeyAlgorithm", "parameters", NULL));
+	params = egg_asn1x_node (asn, "privateKeyAlgorithm", "parameters", NULL);
 
 	ret = SUCCESS;
 
@@ -808,9 +806,7 @@ done:
 	}
 
 	if (keydata)
-		egg_bytes_unref (keydata);
-	if (params)
-		egg_bytes_unref (params);
+		g_bytes_unref (keydata);
 	egg_asn1x_destroy (asn);
 	pop_parsed (self, parsed);
 	return ret;
@@ -818,7 +814,7 @@ done:
 
 static gint
 parse_der_pkcs8_encrypted (GcrParser *self,
-                           EggBytes *data)
+                           GBytes *data)
 {
 	PasswordState pstate = PASSWORD_STATE_INIT;
 	GNode *asn = NULL;
@@ -827,8 +823,8 @@ parse_der_pkcs8_encrypted (GcrParser *self,
 	gint ret, r;
 	GQuark scheme;
 	guchar *crypted = NULL;
-	EggBytes *params = NULL;
-	EggBytes *cbytes;
+	GNode *params = NULL;
+	GBytes *cbytes;
 	gsize n_crypted;
 	const gchar *password;
 	GcrParsed *parsed;
@@ -849,7 +845,7 @@ parse_der_pkcs8_encrypted (GcrParser *self,
 	if (!scheme)
 		goto done;
 
-	params = egg_asn1x_get_element_raw (egg_asn1x_node (asn, "encryptionAlgorithm", "parameters", NULL));
+	params = egg_asn1x_node (asn, "encryptionAlgorithm", "parameters", NULL);
 
 	/* Loop to try different passwords */
 	for (;;) {
@@ -884,13 +880,13 @@ parse_der_pkcs8_encrypted (GcrParser *self,
 		if (l > 0)
 			n_crypted = l;
 
-		cbytes = egg_bytes_new_with_free_func (crypted, n_crypted,
+		cbytes = g_bytes_new_with_free_func (crypted, n_crypted,
 		                                       egg_secure_free, crypted);
 		crypted = NULL;
 
 		/* Try to parse the resulting key */
 		r = parse_der_pkcs8_plain (self, cbytes);
-		egg_bytes_unref (cbytes);
+		g_bytes_unref (cbytes);
 
 		if (r != GCR_ERROR_UNRECOGNIZED) {
 			ret = r;
@@ -901,8 +897,6 @@ parse_der_pkcs8_encrypted (GcrParser *self,
 	}
 
 done:
-	if (params)
-		egg_bytes_unref (params);
 	if (cih)
 		gcry_cipher_close (cih);
 	egg_asn1x_destroy (asn);
@@ -914,7 +908,7 @@ done:
 
 static gint
 parse_der_pkcs8 (GcrParser *self,
-                 EggBytes *data)
+                 GBytes *data)
 {
 	gint ret;
 
@@ -931,7 +925,7 @@ parse_der_pkcs8 (GcrParser *self,
 
 static gint
 parse_der_certificate (GcrParser *self,
-                       EggBytes *data)
+                       GBytes *data)
 {
 	gchar *name = NULL;
 	GcrParsed *parsed;
@@ -977,17 +971,17 @@ parse_der_certificate (GcrParser *self,
 
 static gint
 handle_pkcs7_signed_data (GcrParser *self,
-                          EggBytes *data)
+                          GNode *content)
 {
 	GNode *asn = NULL;
 	GNode *node;
 	gint ret;
-	EggBytes *certificate;
+	GBytes *certificate;
 	int i;
 
 	ret = GCR_ERROR_UNRECOGNIZED;
 
-	asn = egg_asn1x_create_and_decode (pkix_asn1_tab, "pkcs-7-SignedData", data);
+	asn = egg_asn1x_get_any_as (content, pkix_asn1_tab, "pkcs-7-SignedData");
 	if (!asn)
 		goto done;
 
@@ -1001,7 +995,7 @@ handle_pkcs7_signed_data (GcrParser *self,
 
 		certificate = egg_asn1x_get_element_raw (node);
 		ret = parse_der_certificate (self, certificate);
-		egg_bytes_unref (certificate);
+		g_bytes_unref (certificate);
 
 		if (ret != SUCCESS)
 			goto done;
@@ -1018,12 +1012,12 @@ done:
 
 static gint
 parse_der_pkcs7 (GcrParser *self,
-                 EggBytes *data)
+                 GBytes *data)
 {
 	GNode *asn = NULL;
 	GNode *node;
 	gint ret;
-	EggBytes *content = NULL;
+	GNode *content = NULL;
 	GQuark oid;
 	GcrParsed *parsed;
 
@@ -1050,15 +1044,13 @@ parse_der_pkcs7 (GcrParser *self,
 		goto done;
 	}
 
-	content = egg_asn1x_get_element_raw (egg_asn1x_node (asn, "content", NULL));
+	content = egg_asn1x_node (asn, "content", NULL);
 	if (!content)
 		goto done;
 
 	ret = handle_pkcs7_signed_data (self, content);
 
 done:
-	if (content)
-		egg_bytes_unref (content);
 	egg_asn1x_destroy (asn);
 	pop_parsed (self, parsed);
 	return ret;
@@ -1068,63 +1060,31 @@ done:
  * PKCS12
  */
 
-static GNode *
-decode_pkcs12_asn1_accepting_invalid_crap (const EggAsn1xDef *defs,
-                                           const gchar *identifier,
-                                           EggBytes *data)
-{
-	GNode *asn;
-
-	/*
-	 * Because PKCS#12 files, the bags specifically, are notorious for
-	 * being crappily constructed and are often break rules such as DER
-	 * sorting order etc.. we parse the DER in a non-strict fashion.
-	 *
-	 * The rules in DER are designed for X.509 certificates, so there is
-	 * only one way to represent a given certificate (although they fail
-	 * at that as well). But with PKCS#12 we don't have such high
-	 * requirements, and we can slack off on our validation.
-	 */
-
-	asn = egg_asn1x_create (defs, identifier);
-	g_return_val_if_fail (asn != NULL, NULL);
-
-	/* Passing FALSE as the strictness argument */
-	if (!egg_asn1x_decode_no_validate (asn, data) ||
-	    !egg_asn1x_validate (asn, FALSE)) {
-		egg_asn1x_destroy (asn);
-		asn = NULL;
-	}
-
-	return asn;
-}
-
 static gint
 handle_pkcs12_cert_bag (GcrParser *self,
-                        EggBytes *data)
+                        GBytes *data)
 {
 	GNode *asn = NULL;
 	GNode *asn_content = NULL;
 	guchar *certificate = NULL;
-	EggBytes *element = NULL;
+	GNode *element = NULL;
 	gsize n_certificate;
-	EggBytes *bytes;
+	GBytes *bytes;
 	gint ret;
 
 	ret = GCR_ERROR_UNRECOGNIZED;
-	asn = decode_pkcs12_asn1_accepting_invalid_crap (pkix_asn1_tab,
-	                                                 "pkcs-12-CertBag",
-	                                                 data);
+	asn = egg_asn1x_create_and_decode_full (pkix_asn1_tab, "pkcs-12-CertBag",
+	                                        data, EGG_ASN1X_NO_STRICT);
 	if (!asn)
 		goto done;
 
 	ret = GCR_ERROR_FAILURE;
 
-	element = egg_asn1x_get_element_raw (egg_asn1x_node (asn, "certValue", NULL));
+	element = egg_asn1x_node (asn, "certValue", NULL);
 	if (!element)
 		goto done;
 
-	asn_content = egg_asn1x_create_and_decode (pkix_asn1_tab, "pkcs-7-Data", element);
+	asn_content = egg_asn1x_get_any_as (element, pkix_asn1_tab, "pkcs-7-Data");
 	if (!asn_content)
 		goto done;
 
@@ -1132,13 +1092,11 @@ handle_pkcs12_cert_bag (GcrParser *self,
 	if (!certificate)
 		goto done;
 
-	bytes = egg_bytes_new_take (certificate, n_certificate);
+	bytes = g_bytes_new_take (certificate, n_certificate);
 	ret = parse_der_certificate (self, bytes);
-	egg_bytes_unref (bytes);
+	g_bytes_unref (bytes);
 
 done:
-	if (element)
-		egg_bytes_unref (element);
 	egg_asn1x_destroy (asn_content);
 	egg_asn1x_destroy (asn);
 	return ret;
@@ -1150,7 +1108,6 @@ parse_pkcs12_bag_friendly_name (GNode *asn)
 	guint count, i;
 	GQuark oid;
 	GNode *node;
-	EggBytes *element;
 	GNode *asn_str;
 	gchar *result;
 
@@ -1163,10 +1120,7 @@ parse_pkcs12_bag_friendly_name (GNode *asn)
 		if (oid == GCR_OID_PKCS9_ATTRIBUTE_FRIENDLY) {
 			node = egg_asn1x_node (asn, i, "values", 1, NULL);
 			if (node != NULL) {
-				element = egg_asn1x_get_element_raw (node);
-				asn_str = egg_asn1x_create_and_decode (pkix_asn1_tab, "BMPString",
-				                                       element);
-				egg_bytes_unref (element);
+				asn_str = egg_asn1x_get_any_as (node, pkix_asn1_tab, "BMPString");
 				if (asn_str) {
 					result = egg_asn1x_get_bmpstring_as_utf8 (asn_str);
 					egg_asn1x_destroy (asn_str);
@@ -1181,22 +1135,22 @@ parse_pkcs12_bag_friendly_name (GNode *asn)
 
 static gint
 handle_pkcs12_bag (GcrParser *self,
-                   EggBytes *data)
+                   GBytes *data)
 {
 	GNode *asn = NULL;
 	gint ret, r;
 	guint count = 0;
 	GQuark oid;
-	EggBytes *element = NULL;
+	GNode *value;
+	GBytes *element = NULL;
 	gchar *friendly;
 	guint i;
 	GcrParsed *parsed;
 
 	ret = GCR_ERROR_UNRECOGNIZED;
 
-	asn = decode_pkcs12_asn1_accepting_invalid_crap (pkix_asn1_tab,
-	                                                 "pkcs-12-SafeContents",
-	                                                 data);
+	asn = egg_asn1x_create_and_decode_full (pkix_asn1_tab, "pkcs-12-SafeContents",
+	                                        data, EGG_ASN1X_NO_STRICT);
 	if (!asn)
 		goto done;
 
@@ -1215,10 +1169,11 @@ handle_pkcs12_bag (GcrParser *self,
 		if (!oid)
 			goto done;
 
-		element = egg_asn1x_get_element_raw (egg_asn1x_node (asn, i, "bagValue", NULL));
-		if (!element)
+		value = egg_asn1x_node (asn, i, "bagValue", NULL);
+		if (!value)
 			goto done;
 
+		element = egg_asn1x_get_element_raw (value);
 		parsed = push_parsed (self, FALSE);
 
 		friendly = parse_pkcs12_bag_friendly_name (egg_asn1x_node (asn, i, "bagAttributes", NULL));
@@ -1245,7 +1200,7 @@ handle_pkcs12_bag (GcrParser *self,
 		}
 
 		if (element != NULL)
-			egg_bytes_unref (element);
+			g_bytes_unref (element);
 
 		pop_parsed (self, parsed);
 
@@ -1266,26 +1221,25 @@ done:
 
 static gint
 handle_pkcs12_encrypted_bag (GcrParser *self,
-                             EggBytes *data)
+                             GNode *bag)
 {
 	PasswordState pstate = PASSWORD_STATE_INIT;
 	GNode *asn = NULL;
 	gcry_cipher_hd_t cih = NULL;
 	gcry_error_t gcry;
 	guchar *crypted = NULL;
-	EggBytes *params = NULL;
+	GNode *params = NULL;
 	gsize n_crypted;
 	const gchar *password;
-	EggBytes *cbytes;
+	GBytes *cbytes;
 	GQuark scheme;
 	gint ret, r;
 	gint l;
 
 	ret = GCR_ERROR_UNRECOGNIZED;
 
-	asn = decode_pkcs12_asn1_accepting_invalid_crap (pkix_asn1_tab,
-	                                                 "pkcs-7-EncryptedData",
-	                                                 data);
+	asn = egg_asn1x_get_any_as_full (bag, pkix_asn1_tab, "pkcs-7-EncryptedData",
+	                                 EGG_ASN1X_NO_STRICT);
 	if (!asn)
 		goto done;
 
@@ -1296,7 +1250,7 @@ handle_pkcs12_encrypted_bag (GcrParser *self,
 	if (!scheme)
 		goto done;
 
-	params = egg_asn1x_get_element_raw (egg_asn1x_node (asn, "encryptedContentInfo", "contentEncryptionAlgorithm", "parameters", NULL));
+	params = egg_asn1x_node (asn, "encryptedContentInfo", "contentEncryptionAlgorithm", "parameters", NULL);
 	if (!params)
 		goto done;
 
@@ -1336,12 +1290,12 @@ handle_pkcs12_encrypted_bag (GcrParser *self,
 		if (l > 0)
 			n_crypted = l;
 
-		cbytes = egg_bytes_new_with_free_func (crypted, n_crypted, egg_secure_free, crypted);
+		cbytes = g_bytes_new_with_free_func (crypted, n_crypted, egg_secure_free, crypted);
 		crypted = NULL;
 
 		/* Try to parse the resulting key */
 		r = handle_pkcs12_bag (self, cbytes);
-		egg_bytes_unref (cbytes);
+		g_bytes_unref (cbytes);
 
 		if (r != GCR_ERROR_UNRECOGNIZED) {
 			ret = r;
@@ -1352,8 +1306,6 @@ handle_pkcs12_encrypted_bag (GcrParser *self,
 	}
 
 done:
-	if (params)
-		egg_bytes_unref (params);
 	if (cih)
 		gcry_cipher_close (cih);
 	egg_asn1x_destroy (asn);
@@ -1363,22 +1315,21 @@ done:
 
 static gint
 handle_pkcs12_safe (GcrParser *self,
-                    EggBytes *data)
+                    GBytes *data)
 {
 	GNode *asn = NULL;
 	GNode *asn_content = NULL;
 	gint ret, r;
-	EggBytes *bag = NULL;
-	EggBytes *content;
+	GNode *bag;
+	GBytes *content;
 	GQuark oid;
 	guint i;
 	GNode *node;
 
 	ret = GCR_ERROR_UNRECOGNIZED;
 
-	asn = decode_pkcs12_asn1_accepting_invalid_crap (pkix_asn1_tab,
-	                                                 "pkcs-12-AuthenticatedSafe",
-	                                                 data);
+	asn = egg_asn1x_create_and_decode_full (pkix_asn1_tab, "pkcs-12-AuthenticatedSafe",
+	                                        data, EGG_ASN1X_NO_STRICT);
 	if (!asn)
 		goto done;
 
@@ -1396,20 +1347,16 @@ handle_pkcs12_safe (GcrParser *self,
 
 		oid = egg_asn1x_get_oid_as_quark (node);
 
-		node = egg_asn1x_node (asn, i + 1, "content", NULL);
-		if (!node)
+		bag = egg_asn1x_node (asn, i + 1, "content", NULL);
+		if (!bag)
 			goto done;
 
-		bag = egg_asn1x_get_element_raw (node);
-		g_return_val_if_fail (bag != NULL, ret);
-
 		/* A non encrypted bag, just parse */
 		if (oid == GCR_OID_PKCS7_DATA) {
 
 			egg_asn1x_destroy (asn_content);
-			asn_content = decode_pkcs12_asn1_accepting_invalid_crap (pkix_asn1_tab,
-			                                                         "pkcs-7-Data",
-			                                                         bag);
+			asn_content = egg_asn1x_get_any_as_full (bag, pkix_asn1_tab,
+			                                         "pkcs-7-Data", EGG_ASN1X_NO_STRICT);
 			if (!asn_content)
 				goto done;
 
@@ -1418,7 +1365,7 @@ handle_pkcs12_safe (GcrParser *self,
 				goto done;
 
 			r = handle_pkcs12_bag (self, content);
-			egg_bytes_unref (content);
+			g_bytes_unref (content);
 
 		/* Encrypted data first needs decryption */
 		} else if (oid == GCR_OID_PKCS7_ENCRYPTED_DATA) {
@@ -1430,9 +1377,6 @@ handle_pkcs12_safe (GcrParser *self,
 			r = GCR_ERROR_UNRECOGNIZED;
 		}
 
-		egg_bytes_unref (bag);
-		bag = NULL;
-
 		if (r == GCR_ERROR_FAILURE ||
 		    r == GCR_ERROR_CANCELLED ||
 		    r == GCR_ERROR_LOCKED) {
@@ -1444,8 +1388,6 @@ handle_pkcs12_safe (GcrParser *self,
 	ret = SUCCESS;
 
 done:
-	if (bag != NULL)
-		egg_bytes_unref (bag);
 	egg_asn1x_destroy (asn);
 	egg_asn1x_destroy (asn_content);
 	return ret;
@@ -1454,7 +1396,7 @@ done:
 static gint
 verify_pkcs12_safe (GcrParser *self,
                     GNode *asn,
-                    EggBytes *content)
+                    GBytes *content)
 {
 	PasswordState pstate = PASSWORD_STATE_INIT;
 	const gchar *password;
@@ -1465,7 +1407,6 @@ verify_pkcs12_safe (GcrParser *self,
 	gsize n_digest;
 	GQuark algorithm;
 	GNode *mac_data;
-	EggBytes *params = NULL;
 	int ret, r;
 
 	ret = GCR_ERROR_FAILURE;
@@ -1484,10 +1425,6 @@ verify_pkcs12_safe (GcrParser *self,
 	if (!algorithm)
 		goto done;
 
-	params = egg_asn1x_get_element_raw (mac_data);
-	if (!params)
-		goto done;
-
 	digest = egg_asn1x_get_string_as_raw (egg_asn1x_node (mac_data, "mac", "digest", NULL), NULL, &n_digest);
 	if (!digest)
 		goto done;
@@ -1503,7 +1440,7 @@ verify_pkcs12_safe (GcrParser *self,
 		}
 
 		/* Parse the encryption stuff into a cipher. */
-		if (!egg_symkey_read_mac (algorithm, password, -1, params, &mdh, &mac_len)) {
+		if (!egg_symkey_read_mac (algorithm, password, -1, mac_data, &mdh, &mac_len)) {
 			ret = GCR_ERROR_FAILURE;
 			goto done;
 		}
@@ -1513,7 +1450,7 @@ verify_pkcs12_safe (GcrParser *self,
 			r = GCR_ERROR_FAILURE;
 
 		} else {
-			gcry_md_write (mdh, egg_bytes_get_data (content), egg_bytes_get_size (content));
+			gcry_md_write (mdh, g_bytes_get_data (content, NULL), g_bytes_get_size (content));
 			mac_digest = gcry_md_read (mdh, 0);
 			g_return_val_if_fail (mac_digest, GCR_ERROR_FAILURE);
 			r = memcmp (mac_digest, digest, n_digest) == 0 ? SUCCESS : GCR_ERROR_LOCKED;
@@ -1529,8 +1466,6 @@ verify_pkcs12_safe (GcrParser *self,
 	}
 
 done:
-	if (params)
-		egg_bytes_unref (params);
 	if (mdh)
 		gcry_md_close (mdh);
 	g_free (digest);
@@ -1540,20 +1475,31 @@ done:
 
 static gint
 parse_der_pkcs12 (GcrParser *self,
-                  EggBytes *data)
+                  GBytes *data)
 {
 	GNode *asn = NULL;
-	GNode *asn_content = NULL;
 	gint ret;
-	EggBytes *element = NULL;
-	EggBytes *content = NULL;
+	GNode *content = NULL;
+	GBytes *string = NULL;
 	GQuark oid;
 	GcrParsed *parsed;
 
 	parsed = push_parsed (self, FALSE);
 	ret = GCR_ERROR_UNRECOGNIZED;
 
-	asn = decode_pkcs12_asn1_accepting_invalid_crap (pkix_asn1_tab, "pkcs-12-PFX", data);
+	/*
+	 * Because PKCS#12 files, the bags specifically, are notorious for
+	 * being crappily constructed and are often break rules such as DER
+	 * sorting order etc.. we parse the DER in a non-strict fashion.
+	 *
+	 * The rules in DER are designed for X.509 certificates, so there is
+	 * only one way to represent a given certificate (although they fail
+	 * at that as well). But with PKCS#12 we don't have such high
+	 * requirements, and we can slack off on our validation.
+	 */
+
+	asn = egg_asn1x_create_and_decode_full (pkix_asn1_tab, "pkcs-12-PFX",
+	                                        data, EGG_ASN1X_NO_STRICT);
 	if (!asn)
 		goto done;
 
@@ -1569,28 +1515,24 @@ parse_der_pkcs12 (GcrParser *self,
 		goto done;
 	}
 
-	element = egg_asn1x_get_element_raw (egg_asn1x_node (asn, "authSafe", "content", NULL));
-	if (!element)
-		goto done;
-
-	asn_content = decode_pkcs12_asn1_accepting_invalid_crap (pkix_asn1_tab, "pkcs-7-Data", element);
-	if (!asn_content)
-		goto done;
-
-	content = egg_asn1x_get_string_as_bytes (asn_content);
+	content = egg_asn1x_get_any_as (egg_asn1x_node (asn, "authSafe", "content", NULL),
+	                                pkix_asn1_tab, "pkcs-7-Data");
 	if (!content)
 		goto done;
 
-	ret = verify_pkcs12_safe (self, asn, content);
+	string = egg_asn1x_get_string_as_bytes (content);
+	if (!string)
+		goto done;
+
+	ret = verify_pkcs12_safe (self, asn, string);
 	if (ret == SUCCESS)
-		ret = handle_pkcs12_safe (self, content);
+		ret = handle_pkcs12_safe (self, string);
 
 done:
-	if (element)
-		egg_bytes_unref (element);
 	if (content)
-		egg_bytes_unref (content);
-	egg_asn1x_destroy (asn_content);
+		egg_asn1x_destroy (content);
+	if (string)
+		g_bytes_unref (string);
 	egg_asn1x_destroy (asn);
 	pop_parsed (self, parsed);
 	return ret;
@@ -1602,7 +1544,7 @@ done:
 
 static gint
 parse_der_pkcs10 (GcrParser *self,
-                  EggBytes *data)
+                  GBytes *data)
 {
 	GNode *asn = NULL;
 	GNode *node;
@@ -1642,7 +1584,7 @@ parse_der_pkcs10 (GcrParser *self,
 
 static gint
 parse_der_spkac (GcrParser *self,
-                 EggBytes *data)
+                 GBytes *data)
 {
 	GNode *asn = NULL;
 	GcrParsed *parsed;
@@ -1668,7 +1610,7 @@ parse_der_spkac (GcrParser *self,
 
 static gint
 parse_base64_spkac (GcrParser *self,
-                    EggBytes *dat)
+                    GBytes *dat)
 {
 	const gchar *PREFIX = "SPKAC=";
 	const gsize PREFIX_LEN = 6;
@@ -1677,12 +1619,11 @@ parse_base64_spkac (GcrParser *self,
 	guchar *spkac;
 	gsize n_spkac;
 	const guchar *data;
-	EggBytes *bytes;
+	GBytes *bytes;
 	gsize n_data;
 	gint ret;
 
-	data = egg_bytes_get_data (dat);
-	n_data = egg_bytes_get_size (dat);
+	data = g_bytes_get_data (dat, &n_data);
 
 	if (n_data > PREFIX_LEN && memcmp (PREFIX, data, PREFIX_LEN))
 		return GCR_ERROR_UNRECOGNIZED;
@@ -1695,9 +1636,9 @@ parse_base64_spkac (GcrParser *self,
 
 	spkac = g_base64_decode ((const gchar *)data, &n_spkac);
 	if (spkac != NULL) {
-		bytes = egg_bytes_new_take (spkac, n_spkac);
+		bytes = g_bytes_new_take (spkac, n_spkac);
 		ret = parse_der_spkac (self, bytes);
-		egg_bytes_unref (bytes);
+		g_bytes_unref (bytes);
 	} else {
 		ret = GCR_ERROR_FAILURE;
 	}
@@ -1712,7 +1653,7 @@ parse_base64_spkac (GcrParser *self,
 
 static void
 on_openpgp_packet (GPtrArray *records,
-                   EggBytes *outer,
+                   GBytes *outer,
                    gpointer user_data)
 {
 	GcrParser *self = GCR_PARSER (user_data);
@@ -1741,7 +1682,7 @@ on_openpgp_packet (GPtrArray *records,
 
 static gint
 parse_openpgp_packets (GcrParser *self,
-                       EggBytes *data)
+                       GBytes *data)
 {
 	gint num_parsed;
 
@@ -1815,7 +1756,7 @@ static gint
 handle_plain_pem (GcrParser *self,
                   gint format_id,
                   gint want_format,
-                  EggBytes *data)
+                  GBytes *data)
 {
 	ParserFormat *format;
 
@@ -1834,14 +1775,14 @@ handle_encrypted_pem (GcrParser *self,
                       gint format_id,
                       gint want_format,
                       GHashTable *headers,
-                      EggBytes *data)
+                      GBytes *data)
 {
 	PasswordState pstate = PASSWORD_STATE_INIT;
 	const gchar *password;
 	guchar *decrypted;
 	gsize n_decrypted;
 	const gchar *val;
-	EggBytes *dbytes;
+	GBytes *dbytes;
 	gint res;
 	gint l;
 
@@ -1872,13 +1813,13 @@ handle_encrypted_pem (GcrParser *self,
 		if (l > 0)
 			n_decrypted = l;
 
-		dbytes = egg_bytes_new_with_free_func (decrypted, n_decrypted,
+		dbytes = g_bytes_new_with_free_func (decrypted, n_decrypted,
 		                                       egg_secure_free, decrypted);
 		decrypted = NULL;
 
 		/* Try to parse */
 		res = handle_plain_pem (self, format_id, want_format, dbytes);
-		egg_bytes_unref (dbytes);
+		g_bytes_unref (dbytes);
 
 		/* Unrecognized is a bad password */
 		if (res != GCR_ERROR_UNRECOGNIZED)
@@ -1896,8 +1837,8 @@ typedef struct {
 
 static void
 handle_pem_data (GQuark type,
-                 EggBytes *data,
-                 EggBytes *outer,
+                 GBytes *data,
+                 GBytes *outer,
                  GHashTable *headers,
                  gpointer user_data)
 {
@@ -1949,12 +1890,12 @@ handle_pem_data (GQuark type,
 static gint
 handle_pem_format (GcrParser *self,
                    gint subformat,
-                   EggBytes *data)
+                   GBytes *data)
 {
 	HandlePemArgs ctx = { self, GCR_ERROR_UNRECOGNIZED, subformat };
 	guint found;
 
-	if (egg_bytes_get_size (data) == 0)
+	if (g_bytes_get_size (data) == 0)
 		return GCR_ERROR_UNRECOGNIZED;
 
 	found = egg_armor_parse (data, handle_pem_data, &ctx);
@@ -1968,70 +1909,70 @@ handle_pem_format (GcrParser *self,
 
 static gint
 parse_pem (GcrParser *self,
-           EggBytes *data)
+           GBytes *data)
 {
 	return handle_pem_format (self, 0, data);
 }
 
 static gint
 parse_pem_private_key_rsa (GcrParser *self,
-                           EggBytes *data)
+                           GBytes *data)
 {
 	return handle_pem_format (self, GCR_FORMAT_DER_PRIVATE_KEY_RSA, data);
 }
 
 static gint
 parse_pem_private_key_dsa (GcrParser *self,
-                           EggBytes *data)
+                           GBytes *data)
 {
 	return handle_pem_format (self, GCR_FORMAT_DER_PRIVATE_KEY_DSA, data);
 }
 
 static gint
 parse_pem_certificate (GcrParser *self,
-                       EggBytes *data)
+                       GBytes *data)
 {
 	return handle_pem_format (self, GCR_FORMAT_DER_CERTIFICATE_X509, data);
 }
 
 static gint
 parse_pem_pkcs8_plain (GcrParser *self,
-                       EggBytes *data)
+                       GBytes *data)
 {
 	return handle_pem_format (self, GCR_FORMAT_DER_PKCS8_PLAIN, data);
 }
 
 static gint
 parse_pem_pkcs8_encrypted (GcrParser *self,
-                           EggBytes *data)
+                           GBytes *data)
 {
 	return handle_pem_format (self, GCR_FORMAT_DER_PKCS8_ENCRYPTED, data);
 }
 
 static gint
 parse_pem_pkcs7 (GcrParser *self,
-                 EggBytes *data)
+                 GBytes *data)
 {
 	return handle_pem_format (self, GCR_FORMAT_DER_PKCS7, data);
 }
 
 static gint
 parse_pem_pkcs10 (GcrParser *self,
-                  EggBytes *data)
+                  GBytes *data)
 {
 	return handle_pem_format (self, GCR_FORMAT_DER_PKCS10, data);
 }
 
 static gint
 parse_pem_pkcs12 (GcrParser *self,
-                  EggBytes *data)
+                  GBytes *data)
 {
 	return handle_pem_format (self, GCR_FORMAT_DER_PKCS12, data);
 }
 
 static gint
 parse_openpgp_armor (GcrParser *self,
-                     EggBytes *data)
+                     GBytes *data)
 {
 	return handle_pem_format (self, GCR_FORMAT_OPENPGP_PACKET, data);
 }
@@ -2044,7 +1985,7 @@ static void
 on_openssh_public_key_parsed (GckAttributes *attrs,
                               const gchar *label,
                               const gchar *options,
-                              EggBytes *outer,
+                              GBytes *outer,
                               gpointer user_data)
 {
 	GcrParser *self = GCR_PARSER (user_data);
@@ -2060,7 +2001,7 @@ on_openssh_public_key_parsed (GckAttributes *attrs,
 
 static gint
 parse_openssh_public (GcrParser *self,
-                      EggBytes *data)
+                      GBytes *data)
 {
 	guint num_parsed;
 
@@ -2192,7 +2133,7 @@ compare_pointers (gconstpointer a, gconstpointer b)
 
 typedef struct _ForeachArgs {
 	GcrParser *parser;
-	EggBytes *data;
+	GBytes *data;
 	gint result;
 } ForeachArgs;
 
@@ -2354,7 +2295,8 @@ gcr_parser_class_init (GcrParserClass *klass)
 
 	/**
 	 * GcrParser::authenticate:
-	 * @count: The number of times this item has been authenticated.
+	 * @self: the parser
+	 * @count: the number of times this item has been authenticated
 	 *
 	 * This signal is emitted when an item needs to be unlocked or decrypted before
 	 * it can be parsed. The @count argument specifies the number of times
@@ -2377,6 +2319,7 @@ gcr_parser_class_init (GcrParserClass *klass)
 
 	/**
 	 * GcrParser::parsed:
+	 * @self: the parser
 	 *
 	 * This signal is emitted when an item is sucessfully parsed. To access
 	 * the information about the item use the gcr_parser_get_parsed_label(),
@@ -2456,7 +2399,7 @@ gcr_parser_parse_data (GcrParser *self,
 
 	if (data && n_data) {
 		/* TODO: Once GBytes is in the API, copy here */
-		args.data = egg_bytes_new_static (data, n_data);
+		args.data = g_bytes_new_static (data, n_data);
 
 		/* Just the specific formats requested */
 		if (self->pv->specific_formats) {
@@ -2471,7 +2414,7 @@ gcr_parser_parse_data (GcrParser *self,
 			}
 		}
 
-		egg_bytes_unref (args.data);
+		g_bytes_unref (args.data);
 	}
 
 	switch (args.result) {
@@ -2648,7 +2591,7 @@ gcr_parsed_ref (GcrParsed *parsed)
 	/* Find the block of data to copy */
 	while (parsed != NULL) {
 		if (parsed->data != NULL) {
-			copy->data = egg_bytes_ref (parsed->data);
+			copy->data = g_bytes_ref (parsed->data);
 			copy->sensitive = parsed->sensitive;
 			break;
 		}
@@ -2677,7 +2620,7 @@ gcr_parsed_unref (gpointer parsed)
 			gck_attributes_unref (par->attrs);
 		g_free (par->label);
 		if (par->data)
-			egg_bytes_unref (par->data);
+			g_bytes_unref (par->data);
 		g_free (par);
 	}
 }
@@ -2840,10 +2783,8 @@ gcr_parsed_get_data (GcrParsed *parsed,
 	g_return_val_if_fail (n_data != NULL, NULL);
 
 	while (parsed != NULL) {
-		if (parsed->data != NULL) {
-			*n_data = egg_bytes_get_size (parsed->data);
-			return egg_bytes_get_data (parsed->data);
-		}
+		if (parsed->data != NULL)
+			return g_bytes_get_data (parsed->data, n_data);
 		parsed = parsed->next;
 	}
 
